// Mr Watch DCC Monitor
// By yaasan
// Copyright(C) 2017- Desktop Station Co.,Ltd.

#include "NmraDcc_v141.h"

//各種設定、宣言

//#define DEBUG

#define PIN_DCCIN 2

#define DECODER_ADDRESS 3

#define PULSE_MIN	40
#define PULSE_MAX	200

/* DCC SpeedStep 28専用テーブル */
const uint8_t DCC_28SPEEDMAP[29] = {
0b00000,0b00010,0b10010,0b00011,
0b10011,0b00100,0b10100,0b00101,
0b10101,0b00110,0b10110,0b00111,
0b10111,0b01000,0b11000,0b01001,
0b11001,0b01010,0b11010,0b01011,
0b11011,0b01100,0b11100,0b01101,
0b11101,0b01110,0b11110,0b01111,0b11111};

//使用クラスの宣言
NmraDcc	 Dcc;
DCC_MSG	 Packet;

//Task Schedule
unsigned long gPreviousL5 = 0;
unsigned long gPreviousL1 = 0;



uint8_t gButtonStatus = 0;

uint16_t gCount_Err = 0;
uint16_t gCount_Idle = 0;
uint16_t gCount_Reset = 0;
uint16_t gCount_PreMin = 255;
uint16_t gCount_PreMax = 0;


void(* resetFunc) (void) = 0;  //declare reset function at address 0
uint16_t ReadEdcVoltage();

void decodeLocoPacket(uint8_t inPacket1, uint8_t inPacket2);


void setup()
{
	
	//Pin mode
	pinMode(A2, INPUT); //Edc
	pinMode(A3, INPUT); //Button
	digitalWrite(A3, 1); //pullup
	
	
	pinMode(A0, OUTPUT);
	digitalWrite(A0, 0); //Reset High
	delay(50);
	digitalWrite(A0, 1); //Reset High
	delay(50);
	
	//シリアル通信開始
	Serial.begin(115200);

	Serial.println("--------------------------------------");
	Serial.println("Desktop Station Monitor               ");
	Serial.println("--------------------------------------");
	Serial.println("100 Ready");  
	Serial.println("@DSG,999,");

	// Setup which External Interrupt, the Pin it's associated with that we're using, disable pullup.
	Dcc.pin(0, PIN_DCCIN, 0);

	// Call the main DCC Init function to enable the DCC Receiver
	Dcc.init( MAN_ID_DIY, 10, 0 , 0 ); 

	//Reset task
	gPreviousL5 = millis();

}

void loop()
{
	uint8_t aUpdate = 0;
	
	// You MUST call the NmraDcc.process() method frequently from the Arduino loop() function for correct library operation
	Dcc.process();
	
	if( (millis() - gPreviousL1) >= 100)
	{
		
		uint8_t aButton = digitalRead(A3);
		
		//Update Button status
		if( (gButtonStatus == 1) && (aButton == 0))
		{
			//gMode++;
			aUpdate = 1;
			
		}
		
		gButtonStatus = aButton;
		
		//Reset task
		gPreviousL1 = millis();
		
	}
	
	if( ((millis() - gPreviousL5) >= 5000) || (aUpdate == 1))
	{
		//Main Task
		
		Serial.print("(INTERVAL 5sec) RESET:");
		Serial.print(gCount_Reset);
		Serial.print(" IDLE:");
		Serial.print(gCount_Idle);
		Serial.print(" Error:");
		Serial.print(gCount_Err);
		
		Serial.print(" Preamble:");
		Serial.print(gCount_PreMin);
		Serial.print("/");
		Serial.println(gCount_PreMax);
		
		//Clear Counter
		gCount_Reset = 0;
		gCount_Idle = 0;
		gCount_Err = 0;
		
		//線路電圧
		Serial.print("Track Voltage: ");
		Serial.println(ReadEdcVoltage());
		
		
		//Reset task
		gPreviousL5 = millis();
		
	}
}


/* 全DCCメッセージを受信 */
void notifyDccMsg( DCC_MSG * Msg)
{
	uint8_t aErrorbyte = 0;
	uint16_t aLocoAddr = 0;
	uint16_t aAccAddr = 0;
	
	
	
	/* エラーチェック */
	for(uint8_t i = 0; i < Msg->Size; i++)
	{
		aErrorbyte = aErrorbyte ^ Msg->Data[i];
	}
	
	if( aErrorbyte != 0)
	{
		if( gCount_Err <= 65535)
		{
			gCount_Err++;
		}
		
		return;
	}
	else
	{
		//gWatchMode = 1;
	}
	
	/* プリアンブル数計算 */
	if( Msg->PreambleBits < gCount_PreMin)
	{
		gCount_PreMin = Msg->PreambleBits;
	}
	else if( Msg->PreambleBits > gCount_PreMax)
	{
		gCount_PreMax = Msg->PreambleBits;
	}
	
		
	/* パケットチェック */
	switch(Msg->Data[0])
	{
	case 0x00:
		//Broadcast
		
		if( Msg->Data[1] == 0x00)
		{
			gCount_Reset++;
		}
		else
		{
			/* その他のBroadcast処理 */
		}
		
		break;
		
	case 0xFF:
		//Idle packet
		gCount_Idle++;
		break;
		
	default:
		
		if( Msg->Data[0] <= 127)
		{
			/* SHORTアドレスロコの操作 */
			aLocoAddr = Msg->Data[0];
			
			decodeLocoPacket(aLocoAddr, Msg->Data[1], Msg->Data[2]);
			
			
		}
		else if(( Msg->Data[0] >= 128) && ( Msg->Data[0] <= 191))
		{
			/* アクセサリ */
			aAccAddr = (((Msg->Data[0] & 0b111111) - 1) * 4) + ((7 - (Msg->Data[1] >> 4) & 0b111) << 8) +  (((Msg->Data[1] >> 1) & 0b11));
			
			
			if( (Msg->Data[1] & 1) == 0)
			{
				Serial.print("Turnout Adr.");
				Serial.print(aAccAddr + 1);
				Serial.println(" diverse/");
				
				//uiPacketInfo.innerText = uiPacketInfo.innerText + ", direction is diverse.";
			}
			else
			{
				Serial.print("Turnout Adr.");
				Serial.print(aAccAddr + 1);
				Serial.println(" straight|");
				
				//uiPacketInfo.innerText = uiPacketInfo.innerText + ", direction is straight.";
			}
			
		}
		else if(( Msg->Data[0] >= 192) && ( Msg->Data[0] <= 231))
		{
			/* LONGアドレスロコの操作 */
			aLocoAddr = ((Msg->Data[0] - 192) << 8) + Msg->Data[1];
			
			decodeLocoPacket(aLocoAddr, Msg->Data[2], Msg->Data[3]);
		}
		else if(( Msg->Data[0] >= 232) && ( Msg->Data[0] <= 254))
		{
			/* 予約 */
			
		}
		
		
		
		break;
		
	}
	
}

void decodeLocoPacket(word inLocoAddr, uint8_t inPacket1, uint8_t inPacket2)
{
	uint8_t aSpeed;
	uint8_t aDir;
	uint8_t aIndex;
	
	switch(inPacket1 >> 5)
	{
	case 0b000:
		break;
	
	case 0b001:
		
		switch(inPacket1 & 0b11111)
		{
			case 31:
				
				aSpeed = inPacket2 & 0b01111111;
				
				if( aSpeed > 1)
				{
					aSpeed = aSpeed - 1;
				}
				
				/* リスト更新 */
				//updateDccSpeed( inLocoAddr, aSpeed, (inPacket2 >> 7) & 1, 127 );
			
				Serial.print("Loc(127) Adr.");
				Serial.print(inLocoAddr);
				Serial.print(" Spd.");
				Serial.print(aSpeed);
				Serial.println( (((inPacket2 >> 7) & 1) == 0) ? " FWD" : " REV");
			
			break;
		}
		break;

	case 0b010:
		//REV
			
		aSpeed = inPacket1 & 0b11111;
		
		for(int i = 0; i < 29; i++)
		{
			if( aSpeed == DCC_28SPEEDMAP[i])
			{
				/* リスト更新 */
				//updateDccSpeed( inLocoAddr, i, 0, 29 );
				
				Serial.print("Loc(28) Adr.");
				Serial.print(inLocoAddr);
				Serial.print(" Spd.");
				Serial.print(aSpeed);
				Serial.println(" FWD");
				
				break;
			}
		}
		
		break;
	case 0b011:
		//FWD
			
		aSpeed = inPacket1 & 0b11111;
		
		for(int i = 0; i < 29; i++)
		{
			if( aSpeed == DCC_28SPEEDMAP[i])
			{
				/* リスト更新 */
				//updateDccSpeed( inLocoAddr, i, 0, 29 );
			
				Serial.print("Loc(28) Adr.");
				Serial.print(inLocoAddr);
				Serial.print(" Spd.");
				Serial.print(aSpeed);
				Serial.println(" FWD");
				
				break;
			}
		}
		
		break;
		
	case 0b100:
		Serial.print("Loc Adr.");
		Serial.print(inLocoAddr);
		Serial.print(" ");
		Serial.print(((inPacket1 >> 4) & 1) == 0 ? "F0._ " : "F0.O ");
		Serial.print(((inPacket1 >> 0) & 1) == 0 ? "F1._ " : "F1.O ");
		Serial.print(((inPacket1 >> 1) & 1) == 0 ? "F2._ " : "F2.O ");
		Serial.print(((inPacket1 >> 2) & 1) == 0 ? "F3._ " : "F3.O ");
		Serial.print(((inPacket1 >> 3) & 1) == 0 ? "F4._ " : "F4.O ");
		Serial.println("");
		break;
		
	case 0b101:
			
			Serial.print("Loc Adr.");
			Serial.print(inLocoAddr);
			Serial.print(" ");
		
			switch((inPacket1 >> 4) & 1)
			{
			case 1:
				Serial.print(((inPacket1 >> 0) & 1) == 0 ? "F5._ " : "F5.O ");
				Serial.print(((inPacket1 >> 1) & 1) == 0 ? "F6._ " : "F6.O ");
				Serial.print(((inPacket1 >> 2) & 1) == 0 ? "F7._ " : "F7.O ");
				Serial.print(((inPacket1 >> 3) & 1) == 0 ? "F8._ " : "F8.O ");
				break;
			case 0:
				Serial.print(((inPacket1 >> 0) & 1) == 0 ? "F9._ " : "F9.O ");
				Serial.print(((inPacket1 >> 1) & 1) == 0 ? "F10._ " : "F10.O ");
				Serial.print(((inPacket1 >> 2) & 1) == 0 ? "F11._ " : "F11.O ");
				Serial.print(((inPacket1 >> 3) & 1) == 0 ? "F12._ " : "F12.O ");
				break;
			}
			
			Serial.println("");
		break;
		
	case 0b110:
		
		Serial.print("Loc Adr.");
		Serial.print(inLocoAddr);
		Serial.print(" ");
		
		switch(inPacket1 & 0b11111)
		{
		case 30:
				Serial.print(((inPacket2 >> 0) & 1) == 0 ? "F13._ " : "F13.O ");
				Serial.print(((inPacket2 >> 1) & 1) == 0 ? "F14._ " : "F14.O ");
				Serial.print(((inPacket2 >> 2) & 1) == 0 ? "F15._ " : "F15.O ");
				Serial.print(((inPacket2 >> 3) & 1) == 0 ? "F16._ " : "F16.O ");
				Serial.print(((inPacket2 >> 4) & 1) == 0 ? "F17._ " : "F17.O ");
				Serial.print(((inPacket2 >> 5) & 1) == 0 ? "F18._ " : "F18.O ");
				Serial.print(((inPacket2 >> 6) & 1) == 0 ? "F19._ " : "F19.O ");
				Serial.print(((inPacket2 >> 7) & 1) == 0 ? "F20._ " : "F20.O ");
			break;
			
		case 31:
				Serial.print(((inPacket2 >> 0) & 1) == 0 ? "F21._ " : "F21.O ");
				Serial.print(((inPacket2 >> 1) & 1) == 0 ? "F22._ " : "F22.O ");
				Serial.print(((inPacket2 >> 2) & 1) == 0 ? "F23._ " : "F23.O ");
				Serial.print(((inPacket2 >> 3) & 1) == 0 ? "F24._ " : "F24.O ");
				Serial.print(((inPacket2 >> 4) & 1) == 0 ? "F25._ " : "F25.O ");
				Serial.print(((inPacket2 >> 5) & 1) == 0 ? "F26._ " : "F26.O ");
				Serial.print(((inPacket2 >> 6) & 1) == 0 ? "F27._ " : "F27.O ");
				Serial.print(((inPacket2 >> 7) & 1) == 0 ? "F28._ " : "F28.O ");
			break;
		}
		
		Serial.println("");
	}
	

	
}


uint16_t ReadEdcVoltage()
{
	
	//電圧換算
	// 1k / 11k  y=((x*11*5)/1/1024)
	uint16_t aVolt = analogRead(A2);

	//Limit
	if( aVolt > 922)
	{
		//30Vで打ち止め
		aVolt = 922;
	}
	
	aVolt = aVolt >> 2;
	
	aVolt = (aVolt * 301) >> 7;
	
	if( aVolt > 255)
	{
		//UCに入るサイズにリミット(25.5V)
		aVolt = 255;
	}
	
	return aVolt;
	
}



